@@ -1,12 +1,14 @@
+import threading
 import time
 from typing import List
 
 import pytest
 import redis
+from redis import ResponseError
 
 from fakeredis import _msgs as msgs
-from fakeredis._stream import XStream, StreamRangeTest
 from test import testtools
+from test.testtools import resp_conversion, get_protocol_version
 
 
 def get_ids(results):
@@ -20,72 +22,48 @@ def add_items(r: redis.Redis, stream: str, n: int):
     return id_list
 
 
-@pytest.mark.fake
-def test_xstream():
-    stream = XStream()
-    stream.add([0, 0, 1, 1, 2, 2, 3, 3], '0-1')
-    stream.add([1, 1, 2, 2, 3, 3, 4, 4], '1-2')
-    stream.add([2, 2, 3, 3, 4, 4], '1-3')
-    stream.add([3, 3, 4, 4], '2-1')
-    stream.add([3, 3, 4, 4], '2-2')
-    stream.add([3, 3, 4, 4], '3-1')
-    assert stream.add([3, 3, 4, 4], '4-*') == b'4-0'
-    assert stream.last_item_key() == b'4-0'
-    assert stream.add([3, 3, 4, 4], '4-*-*') is None
-    assert len(stream) == 7
-    i = iter(stream)
-    assert next(i) == [b'0-1', [0, 0, 1, 1, 2, 2, 3, 3]]
-    assert next(i) == [b'1-2', [1, 1, 2, 2, 3, 3, 4, 4]]
-    assert next(i) == [b'1-3', [2, 2, 3, 3, 4, 4]]
-    assert next(i) == [b'2-1', [3, 3, 4, 4]]
-    assert next(i) == [b'2-2', [3, 3, 4, 4]]
-
-    assert stream.find_index_key_as_str('1-2') == (1, True)
-    assert stream.find_index_key_as_str('0-1') == (0, True)
-    assert stream.find_index_key_as_str('2-1') == (3, True)
-    assert stream.find_index_key_as_str('1-4') == (3, False)
-
-    lst = stream.irange(StreamRangeTest.decode(b'0-2'), StreamRangeTest.decode(b'3-0'))
-    assert len(lst) == 4
-
-    stream = XStream()
-    assert stream.delete(['1']) == 0
-    entry_key: bytes = stream.add([0, 0, 1, 1, 2, 2, 3, 3])
-    assert len(stream) == 1
-    assert stream.delete([entry_key, ]) == 1
-    assert len(stream) == 0
+@pytest.mark.resp3_only
+@pytest.mark.decode_responses
+def test_xadd_xread_resp3(r: redis.Redis):
+    stream = "stream"
+    fields = {"some": "other"}
+    m1 = r.xadd(stream, fields=fields)
+    res = r.xread({stream: "0-0"}, block=0)
+    assert isinstance(res, dict)
+    assert stream in res
+    assert res[stream] == [[(m1, fields)]]
 
 
 def test_xadd_redis__green(r: redis.Redis):
     stream = "stream"
     before = int(1000 * time.time())
     m1 = r.xadd(stream, {"some": "other"})
+    ts1, seq1 = m1.decode().split("-")
     after = int(1000 * time.time()) + 1
-    ts1, seq1 = m1.decode().split('-')
     assert before <= int(ts1) <= after
     seq1 = int(seq1)
-    m2 = r.xadd(stream, {'add': 'more'}, id=f'{ts1}-{seq1 + 1}')
-    ts2, seq2 = m2.decode().split('-')
+    m2 = r.xadd(stream, {"add": "more"}, id=f"{ts1}-{seq1 + 1}")
+    ts2, seq2 = m2.decode().split("-")
     assert ts1 == ts2
     assert int(seq2) == int(seq1) + 1
 
     stream = "stream2"
     m1 = r.xadd(stream, {"some": "other"})
-    ts1, seq1 = m1.decode().split('-')
+    ts1, seq1 = m1.decode().split("-")
     ts1 = int(ts1) - 1
     with pytest.raises(redis.ResponseError):
-        r.xadd(stream, {'add': 'more'}, id=f'{ts1}-*')
+        r.xadd(stream, {"add": "more"}, id=f"{ts1}-*")
     with pytest.raises(redis.ResponseError):
-        r.xadd(stream, {'add': 'more'}, id=f'{ts1}-1')
+        r.xadd(stream, {"add": "more"}, id=f"{ts1}-1")
 
 
-@pytest.mark.min_server('7')
+@pytest.mark.min_server("7")
 def test_xadd_redis7(r: redis.Redis):  # Using ts-*
     stream = "stream"
     m1 = r.xadd(stream, {"some": "other"})
-    ts1, seq1 = m1.decode().split('-')
-    m2 = r.xadd(stream, {'add': 'more'}, id=f'{ts1}-*')
-    ts2, seq2 = m2.decode().split('-')
+    ts1, seq1 = m1.decode().split("-")
+    m2 = r.xadd(stream, {"add": "more"}, id=f"{ts1}-*")
+    ts2, seq2 = m2.decode().split("-")
     ts1, seq1 = int(ts1), int(seq1)
     ts2, seq2 = int(ts2), int(seq2)
     assert ts2 == ts1
@@ -96,24 +74,22 @@ def test_xadd_maxlen(r: redis.Redis):
     stream = "stream"
     id_list = add_items(r, stream, 10)
     maxlen = 5
-    id_list.append(r.xadd(stream, {'k': 'new'}, maxlen=maxlen, approximate=False))
+    id_list.append(r.xadd(stream, {"k": "new"}, maxlen=maxlen, approximate=False))
     assert r.xlen(stream) == maxlen
     results = r.xrange(stream, id_list[0])
-    assert get_ids(results) == id_list[len(id_list) - maxlen:]
+    assert get_ids(results) == id_list[len(id_list) - maxlen :]
     with pytest.raises(redis.ResponseError):
-        testtools.raw_command(
-            r, 'xadd', stream,
-            'maxlen', '3', 'minid', 'sometestvalue', 'field', 'value')
-    assert r.set('non-a-stream', 1) == 1
+        testtools.raw_command(r, "xadd", stream, "maxlen", "3", "minid", "sometestvalue", "field", "value")
+    assert r.set("non-a-stream", 1) == 1
     with pytest.raises(redis.ResponseError):
-        r.xlen('non-a-stream')
+        r.xlen("non-a-stream")
 
 
 def test_xadd_minid(r: redis.Redis):
     stream = "stream"
     id_list = add_items(r, stream, 10)
     minid = id_list[6]
-    id_list.append(r.xadd(stream, {'k': 'new'}, minid=minid, approximate=False))
+    id_list.append(r.xadd(stream, {"k": "new"}, minid=minid, approximate=False))
     assert r.xlen(stream) == len(id_list) - 6
     results = r.xrange(stream, id_list[0])
     assert get_ids(results) == id_list[6:]
@@ -146,9 +122,9 @@ def test_xtrim_minlen_and_length_args(r: redis.Redis):
         assert r.xtrim(stream, maxlen=3, minid="sometestvalue")
 
     with pytest.raises(redis.ResponseError):
-        testtools.raw_command(r, 'xtrim', stream, 'maxlen', '3', 'minid', 'sometestvalue')
+        testtools.raw_command(r, "xtrim", stream, "maxlen", "3", "minid", "sometestvalue")
     # minid with a limit
-    stream = 's2'
+    stream = "s2"
     m1 = add_items(r, stream, 4)[0]
     assert r.xtrim(stream, minid=m1, limit=3) == 0
 
@@ -165,8 +141,8 @@ def test_xtrim_minlen_and_length_args(r: redis.Redis):
 
 
 def test_xadd_nomkstream(r: redis.Redis):
-    r.xadd('stream2', {"some": "other"}, nomkstream=True)
-    assert r.xlen('stream2') == 0
+    r.xadd("stream2", {"some": "other"}, nomkstream=True)
+    assert r.xlen("stream2") == 0
     # nomkstream option
     stream = "stream"
     r.xadd(stream, {"foo": "bar"})
@@ -200,14 +176,34 @@ def test_xrevrange(r: redis.Redis):
     assert get_ids(results) == [m4]
 
 
+def test_xrevrange_exclusive(r: redis.Redis):
+    stream = "stream"
+    m1, m2, m3, m4 = _add_to_stream(r, stream, 4)
+
+    def _exc(key: bytes) -> bytes:
+        return b"(" + key
+
+    results = r.xrevrange(stream, max=_exc(m4))
+    assert get_ids(results) == [m3, m2, m1]
+
+    results = r.xrevrange(stream, max=_exc(m3), min=m2)
+    assert get_ids(results) == [m2]
+
+    results = r.xrevrange(stream, min=_exc(m3))
+    assert get_ids(results) == [m4]
+
+    results = r.xrevrange(stream, min=_exc(m1))
+    assert get_ids(results) == [m4, m3, m2]
+
+
 def test_xrange(r: redis.Redis):
-    m = r.xadd('stream1', {"foo": "bar"})
-    assert r.xrange('stream1') == [(m, {b"foo": b"bar"}), ]
+    m = r.xadd("stream1", {"foo": "bar"})
+    assert r.xrange("stream1") == [(m, {b"foo": b"bar"})]
 
-    stream = 'stream2'
-    m = testtools.raw_command(r, 'xadd', stream, '*', b'field', b'value', b'foo', b'bar')
+    stream = "stream2"
+    m = testtools.raw_command(r, "xadd", stream, "*", b"field", b"value", b"foo", b"bar")
 
-    assert r.xrevrange(stream) == [(m, {b'field': b'value', b'foo': b'bar'}), ]
+    assert r.xrevrange(stream) == [(m, {b"field": b"value", b"foo": b"bar"})]
 
     stream = "stream"
     m1, m2, m3, m4 = _add_to_stream(r, stream, 4)
@@ -232,36 +228,81 @@ def get_stream_message(client, stream, message_id):
     return response[0]
 
 
+def test_xread_multiple_streams_blocking(r: redis.Redis):
+    stream1 = "stream1"
+    stream2 = "stream2"
+    r.xadd(stream1, {"foo": "bar"})
+    r.xadd(stream2, {"bing": "baz"})
+
+    res = r.xread(streams={stream1: 0, stream2: 0}, block=10)
+    assert len(res) == 2
+
+
+def test_xread_blocking_no_count(r: redis.Redis):
+    k = b"key"
+    r.xadd(k, {"value": 1234})
+    streams = {k: "0"}
+    m1 = r.xread(streams=streams, block=10)
+    if get_protocol_version(r) == 3:
+        m1 = m1[k][0]
+    else:
+        m1 = m1[0][1]
+    assert m1[0][1] == {b"value": b"1234"}
+
+
 def test_xread(r: redis.Redis):
     stream = "stream"
     m1 = r.xadd(stream, {"foo": "bar"})
     m2 = r.xadd(stream, {"bing": "baz"})
 
-    expected = [[
-        stream.encode(),
-        [get_stream_message(r, stream, m1), get_stream_message(r, stream, m2)],
-    ]]
     # xread starting at 0 returns both messages
-    assert r.xread(streams={stream: 0}) == expected
+    actual = r.xread(streams={stream: 0})
+    assert actual == resp_conversion(
+        r,
+        {stream.encode(): [[get_stream_message(r, stream, m1), get_stream_message(r, stream, m2)]]},
+        [[stream.encode(), [get_stream_message(r, stream, m1), get_stream_message(r, stream, m2)]]],
+    )
 
-    expected = [[stream.encode(), [get_stream_message(r, stream, m1)]]]
     # xread starting at 0 and count=1 returns only the first message
-    assert r.xread(streams={stream: 0}, count=1) == expected
+    assert r.xread(streams={stream: 0}, count=1) == resp_conversion(
+        r,
+        {stream.encode(): [[get_stream_message(r, stream, m1)]]},
+        [[stream.encode(), [get_stream_message(r, stream, m1)]]],
+    )
 
-    expected = [[stream.encode(), [get_stream_message(r, stream, m2)]]]
     # xread starting at m1 returns only the second message
-    assert r.xread(streams={stream: m1}) == expected
+    assert r.xread(streams={stream: m1}) == resp_conversion(
+        r,
+        {stream.encode(): [[get_stream_message(r, stream, m2)]]},
+        [[stream.encode(), [get_stream_message(r, stream, m2)]]],
+    )
 
     # xread starting at the last message returns an empty list
-    assert r.xread(streams={stream: m2}) == []
+    assert r.xread(streams={stream: m2}) == resp_conversion(r, {}, [])
+
+
+def test_xread_count(r: redis.Redis):
+    r.xadd("test", {"x": 1})
+    result = r.xread(streams={"test": "0"}, count=100, block=10)
+    if get_protocol_version(r) == 3:
+        key, values = b"test", result[b"test"][0][0][1]
+    else:
+        key, values = result[0][0], result[0][1][0][1]
+    assert key == b"test"
+    assert values == {b"x": b"1"}
 
 
 def test_xread_bad_commands(r: redis.Redis):
     with pytest.raises(redis.ResponseError) as exc_info:
-        testtools.raw_command(r, 'xread', 'foo', '11-1')
+        testtools.raw_command(r, "xread", "foo", "11-1")
     print(exc_info)
     with pytest.raises(redis.ResponseError) as ex2:
-        testtools.raw_command(r, 'xread', 'streams', 'foo', )
+        testtools.raw_command(
+            r,
+            "xread",
+            "streams",
+            "foo",
+        )
     print(ex2)
 
 
@@ -280,9 +321,9 @@ def test_xdel(r: redis.Redis):
     assert r.xdel(stream, m2, m3) == 2
 
     with pytest.raises(redis.ResponseError) as ex:
-        testtools.raw_command(r, 'XDEL', stream)
-    assert ex.value.args[0] == msgs.WRONG_ARGS_MSG6.format('xdel')[4:]
-    assert r.xdel('non-existing-key', '1-1') == 0
+        testtools.raw_command(r, "XDEL", stream)
+    assert ex.value.args[0] == msgs.WRONG_ARGS_MSG6.format("xdel")[4:]
+    assert r.xdel("non-existing-key", "1-1") == 0
 
 
 def test_xgroup_destroy(r: redis.Redis):
@@ -296,38 +337,40 @@ def test_xgroup_destroy(r: redis.Redis):
     assert r.xgroup_destroy(stream, group) == 1
 
 
-@pytest.mark.max_server('6.3')
-def test_xgroup_create_redis6(r: redis.Redis):
+@pytest.mark.max_server("6.3")
+def test_xgroup_create_connection6(r: redis.Redis):
     stream, group = "stream", "group"
     message_id = r.xadd(stream, {"foo": "bar"})
     r.xgroup_create(stream, group, message_id)
     r.xadd(stream, {"foo": "bar"})
     res = r.xinfo_groups(stream)
     assert len(res) == 1
-    assert res[0]['name'] == group.encode()
-    assert res[0]['consumers'] == 0
-    assert res[0]['pending'] == 0
-    assert res[0]['last-delivered-id'] == message_id
+    assert res[0]["name"] == group.encode()
+    assert res[0]["consumers"] == 0
+    assert res[0]["pending"] == 0
+    assert res[0]["last-delivered-id"] == message_id
 
 
-@pytest.mark.min_server('7')
-def test_xgroup_create_redis7(r: redis.Redis):
+@pytest.mark.min_server("7")
+def test_xgroup_create_connection7(r: redis.Redis):
     stream, group = "stream", "group"
     message_id = r.xadd(stream, {"foo": "bar"})
     r.xgroup_create(stream, group, message_id)
     r.xadd(stream, {"foo": "bar"})
-    expected = [{
-        "name": group.encode(),
-        "consumers": 0,
-        "pending": 0,
-        "last-delivered-id": message_id,
-        "entries-read": None,
-        "lag": 1,
-    }]
+    expected = [
+        {
+            "name": group.encode(),
+            "consumers": 0,
+            "pending": 0,
+            "last-delivered-id": message_id,
+            "entries-read": None,
+            "lag": 1,
+        }
+    ]
     assert r.xinfo_groups(stream) == expected
 
 
-@pytest.mark.min_server('7')
+@pytest.mark.min_server("7")
 def test_xgroup_setid_redis7(r: redis.Redis):
     stream, group = "stream", "group"
     message_id = r.xadd(stream, {"foo": "bar"})
@@ -335,14 +378,16 @@ def test_xgroup_setid_redis7(r: redis.Redis):
     r.xgroup_create(stream, group, 0)
     # advance the last_delivered_id to the message_id
     r.xgroup_setid(stream, group, message_id, entries_read=2)
-    expected = [{
-        "name": group.encode(),
-        "consumers": 0,
-        "pending": 0,
-        "last-delivered-id": message_id,
-        "entries-read": 2,
-        "lag": -1,
-    }]
+    expected = [
+        {
+            "name": group.encode(),
+            "consumers": 0,
+            "pending": 0,
+            "last-delivered-id": message_id,
+            "entries-read": 2,
+            "lag": -1,
+        }
+    ]
     assert r.xinfo_groups(stream) == expected
 
 
@@ -406,23 +451,35 @@ def test_xreadgroup(r: redis.Redis):
     stream, group, consumer = "stream", "group", "consumer1"
     with pytest.raises(redis.exceptions.ResponseError):
         r.xreadgroup(group, consumer, streams={stream: ">"})
-    m1 = r.xadd(stream, {"foo": "bar"})
-    m2 = r.xadd(stream, {"bing": "baz"})
+    c1 = {b"foo": b"bar"}
+    c2 = {b"bing": b"baz"}
+    m1 = r.xadd(stream, c1)
+    m2 = r.xadd(stream, c2)
+    with pytest.raises(
+        redis.exceptions.ResponseError, match=msgs.XREADGROUP_KEY_OR_GROUP_NOT_FOUND_MSG.format(stream, group)
+    ):
+        r.xreadgroup(group, consumer, streams={stream: ">"})
     r.xgroup_create(stream, group, 0)
 
-    expected = [[
-        stream.encode(),
-        [get_stream_message(r, stream, m1), get_stream_message(r, stream, m2)],
-    ]]
+    expected_resp2 = [
+        [
+            stream.encode(),
+            [get_stream_message(r, stream, m1), get_stream_message(r, stream, m2)],
+        ]
+    ]
+    expected_resp3 = {stream.encode(): [[get_stream_message(r, stream, m1), get_stream_message(r, stream, m2)]]}
+    actual = r.xreadgroup(group, consumer, streams={stream: ">"})
     # xread starting at 0 returns both messages
-    assert r.xreadgroup(group, consumer, streams={stream: ">"}) == expected
+    assert actual == resp_conversion(r, expected_resp3, expected_resp2)
 
     r.xgroup_destroy(stream, group)
     r.xgroup_create(stream, group, 0)
 
-    expected = [[stream.encode(), [get_stream_message(r, stream, m1)]]]
+    expected_resp2 = [[stream.encode(), [get_stream_message(r, stream, m1)]]]
+    expected_resp3 = {stream.encode(): [[get_stream_message(r, stream, m1)]]}
+    actual = r.xreadgroup(group, consumer, streams={stream: ">"}, count=1)
     # xread with count=1 returns only the first message
-    assert r.xreadgroup(group, consumer, streams={stream: ">"}, count=1) == expected
+    assert actual == resp_conversion(r, expected_resp3, expected_resp2)
 
     r.xgroup_destroy(stream, group)
 
@@ -430,27 +487,39 @@ def test_xreadgroup(r: redis.Redis):
     # will only find messages added after this
     r.xgroup_create(stream, group, "$")
 
-    expected = []
     # xread starting after the last message returns an empty message list
-    assert r.xreadgroup(group, consumer, streams={stream: ">"}) == expected
+    assert r.xreadgroup(group, consumer, streams={stream: ">"}) == resp_conversion(r, {}, [])
 
     # xreadgroup with noack does not have any items in the PEL
     r.xgroup_destroy(stream, group)
     r.xgroup_create(stream, group, "0")
-    assert len(r.xreadgroup(group, consumer, streams={stream: ">"}, noack=True)[0][1]) == 2
+    tmp = r.xreadgroup(group, consumer, streams={stream: ">"}, noack=True)
+    if get_protocol_version(r) == 3:
+        res = tmp[stream.encode()][0]
+    else:
+        res = tmp[0][1]
+    assert len(res) == 2
     # now there should be nothing pending
-    res = r.xreadgroup(group, consumer, streams={stream: "0"})
-    assert len(res[0][1]) == 0
+    tmp = r.xreadgroup(group, consumer, streams={stream: "0"})
+    if get_protocol_version(r) == 3:
+        res = tmp[stream.encode()][0]
+    else:
+        res = tmp[0][1]
+    assert len(res) == 0
 
     r.xgroup_destroy(stream, group)
     r.xgroup_create(stream, group, "0")
 
-    # TODO
+    expected_resp2 = [[stream.encode(), [(m1, c1), (m2, c2)]]]
+    expected_resp3 = {stream.encode(): [[(m1, c1), (m2, c2)]]}
+    actual = r.xreadgroup(group, consumer, streams={stream: ">"})
+    assert actual == resp_conversion(r, expected_resp3, expected_resp2)
     # delete all the messages in the stream
+    assert r.xtrim(stream, 0) == 2
+    # TODO groups keep ids of deleted messages
     # expected = [[stream.encode(), [(m1, {}), (m2, {})]]]
-    # r.xreadgroup(group, consumer, streams={stream: ">"})
-    # r.xtrim(stream, 0)
     # assert r.xreadgroup(group, consumer, streams={stream: "0"}) == expected
+    r.xreadgroup(group, consumer, streams={stream: ">"}, count=10, block=500)
 
 
 def test_xinfo_stream(r: redis.Redis):
@@ -464,16 +533,15 @@ def test_xinfo_stream(r: redis.Redis):
     assert info["last-entry"] == get_stream_message(r, stream, m2)
 
 
-def assert_consumer_info(
-        r: redis.Redis, stream: str, group: str, equal_keys: List) -> List:
+def assert_consumer_info(r: redis.Redis, stream: str, group: str, equal_keys: List) -> List:
     res = r.xinfo_consumers(stream, group)
     assert len(res) == len(equal_keys)
     for i in range(len(equal_keys)):
         for k in res[i]:
             if k in equal_keys[i]:
-                assert res[i][k] == equal_keys[i][k], f'res[{i}][{k}] mismatch, {res}!={equal_keys}'
+                assert res[i][k] == equal_keys[i][k], f"res[{i}][{k}] mismatch, {res}!={equal_keys}"
             else:
-                print(f'res[{i}][{k}]={res[i][k]}')
+                print(f"res[{i}][{k}]={res[i][k]}")
     return res
 
 
@@ -491,16 +559,16 @@ def test_xack(r: redis.Redis):
 
     r.xgroup_create(stream, group, 0)
     r.xreadgroup(group, consumer, streams={stream: ">"})
-    assert_consumer_info(r, stream, group, [{'name': b'consumer', 'pending': 3}])
+    assert_consumer_info(r, stream, group, [{"name": b"consumer", "pending": 3}])
     assert r.xack(stream, group, m1) == 1
     time.sleep(0.01)
-    res = assert_consumer_info(r, stream, group, [{'name': b'consumer', 'pending': 2}])
-    assert 'idle' in res[0] and res[0]['idle'] > 0
+    res = assert_consumer_info(r, stream, group, [{"name": b"consumer", "pending": 2}])
+    assert "idle" in res[0] and res[0]["idle"] > 0
     assert r.xack(stream, group, m2, m3) == 2
-    assert_consumer_info(r, stream, group, [{'name': b'consumer', 'pending': 0}])
+    assert_consumer_info(r, stream, group, [{"name": b"consumer", "pending": 0}])
 
 
-@pytest.mark.min_server('7')
+@pytest.mark.min_server("7")
 def test_xinfo_stream_redis7(r: redis.Redis):
     stream = "stream"
     m1 = r.xadd(stream, {"foo": "bar"})
@@ -525,6 +593,9 @@ def test_xinfo_stream_redis7(r: redis.Redis):
     assert info["entries-added"] == 2
     assert info["recorded-first-entry-id"] == b"0-0"
 
+    with pytest.raises(redis.exceptions.ResponseError):
+        r.xinfo_stream("non-existing-key")
+
 
 def test_xinfo_stream_full(r: redis.Redis):
     stream, group = "stream", "group"
@@ -625,36 +696,8 @@ def test_xpending_range_negative(r: redis.Redis):
         r.xpending_range(stream, group, min=None, max=None, count=None, consumername=0)
 
 
-@pytest.mark.max_server('6.3')
-@testtools.run_test_if_redispy_ver('above', '4.4')
-def test_xautoclaim_redis6(r: redis.Redis):
-    stream, group, consumer1, consumer2 = "stream", "group", "consumer1", "consumer2"
-
-    message_id1 = r.xadd(stream, {"john": "wick"})
-    message_id2 = r.xadd(stream, {"johny": "deff"})
-    message = get_stream_message(r, stream, message_id1)
-    r.xgroup_create(stream, group, 0)
-
-    # trying to claim a message that isn't already pending doesn't
-    # do anything
-    assert r.xautoclaim(stream, group, consumer2, min_idle_time=0) == [b"0-0", []]
-
-    # read the group as consumer1 to initially claim the messages
-    r.xreadgroup(group, consumer1, streams={stream: ">"})
-
-    # claim one message as consumer2
-    response = r.xautoclaim(stream, group, consumer2, min_idle_time=0, count=1)
-    assert response[1] == [message]
-
-    # reclaim the messages as consumer1, but use the justid argument
-    # which only returns message ids
-    assert r.xautoclaim(stream, group, consumer1, min_idle_time=0, start_id=0, justid=True) == [
-        message_id1, message_id2]
-    assert r.xautoclaim(stream, group, consumer1, min_idle_time=0, start_id=message_id2, justid=True) == [message_id2]
-
-
-@pytest.mark.min_server('7')
-@testtools.run_test_if_redispy_ver('above', '4.4')
+@pytest.mark.min_server("7")
+@testtools.run_test_if_redispy_ver("gte", "4.4")
 def test_xautoclaim_redis7(r: redis.Redis):
     stream, group, consumer1, consumer2 = "stream", "group", "consumer1", "consumer2"
 
@@ -677,11 +720,13 @@ def test_xautoclaim_redis7(r: redis.Redis):
     # reclaim the messages as consumer1, but use the justid argument
     # which only returns message ids
     assert r.xautoclaim(stream, group, consumer1, min_idle_time=0, start_id=0, justid=True) == [
-        message_id1, message_id2]
+        message_id1,
+        message_id2,
+    ]
     assert r.xautoclaim(stream, group, consumer1, min_idle_time=0, start_id=message_id2, justid=True) == [message_id2]
 
 
-@pytest.mark.min_server('7')
+@pytest.mark.min_server("7")
 def test_xclaim_trimmed_redis7(r: redis.Redis):
     # xclaim should not raise an exception if the item is not there
     stream, group = "stream", "group"
@@ -714,21 +759,88 @@ def test_xclaim(r: redis.Redis):
 
     # trying to claim a message that isn't already pending doesn't
     # do anything
-    assert r.xclaim(
-        stream, group, consumer2, min_idle_time=0, message_ids=(message_id,)
-    ) == []
+    assert r.xclaim(stream, group, consumer2, min_idle_time=0, message_ids=(message_id,)) == []
 
     # read the group as consumer1 to initially claim the messages
     r.xreadgroup(group, consumer1, streams={stream: ">"})
 
     # claim the message as consumer2
-    assert r.xclaim(
-        stream, group, consumer2, min_idle_time=0, message_ids=(message_id,)
-    ) == [message, ]
+    assert r.xclaim(stream, group, consumer2, min_idle_time=0, message_ids=(message_id,)) == [message]
 
     # reclaim the message as consumer1, but use the justid argument
     # which only returns message ids
-    assert r.xclaim(
-        stream, group, consumer1,
-        min_idle_time=0, message_ids=(message_id,), justid=True,
-    ) == [message_id, ]
+    assert r.xclaim(stream, group, consumer1, min_idle_time=0, message_ids=(message_id,), justid=True) == [message_id]
+
+
+def test_xread_blocking(create_connection):
+    # thread with xread block 0 should hang
+    # putting data in the stream should unblock it
+    event = threading.Event()
+    event.clear()
+
+    def thread_func():
+        while not event.is_set():
+            time.sleep(0.1)
+        r = create_connection(db=1)
+        r.xadd("stream", {"x": "1"})
+        time.sleep(0.1)
+
+    t = threading.Thread(target=thread_func)
+    t.start()
+    r1 = create_connection(db=1)
+    event.set()
+    result = r1.xread({"stream": "$"}, block=0, count=1)
+    event.clear()
+    t.join()
+    if get_protocol_version(r1) == 2:
+        assert result[0][0] == b"stream"
+        assert result[0][1][0][1] == {b"x": b"1"}
+    else:
+        assert b"stream" in result
+        assert result[b"stream"][0][0][1] == {b"x": b"1"}
+
+
+def test_stream_ttl(r: redis.Redis):
+    stream = "stream"
+
+    m1 = r.xadd(stream, {"foo": "bar"})
+    assert r.xread(streams={stream: 0}) == resp_conversion(
+        r, {b"stream": [[get_stream_message(r, stream, m1)]]}, [[stream.encode(), [get_stream_message(r, stream, m1)]]]
+    )
+    assert r.xtrim(stream, 0) == 1
+    assert r.ttl(stream) == -1
+
+
+def test_xreadgroup_length_less_than_count(r: redis.Redis):
+    r.xadd("test-events", {"message": "hello"})
+    r.xadd("test-events", {"message": "bye"})
+
+    r.xgroup_create("test-events", "group1", id="0", mkstream=True)
+    messages = r.xreadgroup(
+        groupname="group1", consumername="consumer1", streams={"test-events": ">"}, count=10, block=2000
+    )
+    assert len(messages) == 1
+
+
+def test_xadd_change_time(r: redis.Redis):
+    res = r.xadd("foobar", {"a": "1"})
+    ts, seq = res.decode().split("-")
+    new_ts = int(ts) - 10
+    new_id = f"{new_ts}-*"
+    with pytest.raises(ResponseError):
+        r.xadd("foobar", {"a": "2"}, id=new_id)
+
+
+def test_xinfo_groups_pending(r: redis.Redis):
+    stream_name = "testing_stream"
+    group_name = "testing_group"
+    consumer_name = "testing_consumer"
+
+    r.delete(stream_name)
+    r.xgroup_create(stream_name, group_name, id="0", mkstream=True)
+
+    r.xadd(stream_name, {"key": "value"})
+
+    r.xreadgroup(group_name, consumer_name, {stream_name: ">"}, count=1)
+    assert r.xpending(stream_name, group_name)["pending"] == 1
+    assert r.xinfo_groups(stream_name)[0]["pending"] == 1
